/* Expand one page of G3 fax data
   Copyright (C) 1990, 1995  Frank D. Cringle.

This file is part of g3view - an X11 fax viewer
     
g3view is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2 of the License, or (at your
option) any later version.
     
This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.
     
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include <stdlib.h>
#include <stdio.h>
#include <X11/Xlib.h>
#include "g3.h"

#undef BITS_LOW
/* Here are 2 logically equivalent stategies for cacheing bits in a 32
   bit accumulator: left-aligned or right-aligned.  Choose whichever
   is faster on your machine.  Weep if your CPU does not have a shift
   matrix and does shifts 1 bit at a time (e.g. 68K).
   Speed addicts can win by using asm() here (e.g. 88K extu). */
#ifdef BITS_LOW
#define NeedBits(n)							\
    while (BitsAvail < (n)) {						\
	BitAcc = (BitAcc << 8) | *sp++;					\
	BitsAvail += 8;							\
    }
#define GetBits(n)	(BitAcc >> (BitsAvail - (n)))
#define ClrBits(n)							\
    do {								\
	BitsAvail -= (n);						\
	BitAcc &= ((1<<BitsAvail)-1);					\
    } while (0)
#define DEBUG_SHOW putchar(BitAcc & (1 << (BitsAvail-1-t)) ? '1' : '0')
#else  /* BITS_HIGH */
#define NeedBits(n)							\
    while (BitsAvail < (n)) {						\
	BitAcc |= (*sp++ << (24 - BitsAvail));				\
	BitsAvail += 8;							\
    }
#define GetBits(n)	(BitAcc >> (32 - (n)))
#define ClrBits(n)							\
    do {								\
	BitsAvail -= (n);						\
	BitAcc <<= (n);							\
    } while (0)
#define DEBUG_SHOW putchar(BitAcc & (1 << (31-t)) ? '1' : '0')
#endif
#ifdef DEBUG
#define LOOKUP(wid,tab)	{						\
    int t;								\
    NeedBits(wid);							\
    TabEnt = tab + GetBits(wid);					\
    printf("%08lX/%d: %s%5d\t", BitAcc, BitsAvail,			\
	   StateNames[TabEnt->State], TabEnt->Param);			\
    for (t = 0; t < TabEnt->Width; t++)					\
	DEBUG_SHOW;							\
    putchar('\n');							\
    ClrBits(TabEnt->Width);						\
}

#define SETVAL(x,c) {							\
    *pa++ = RunLength + (x);						\
    printf("SETVAL: %s(%d)\t%d\n",					\
	   RunColour ? "black" : "white", RunLength + (x), a0);		\
    a0 += x;								\
    RunLength = 0;							\
    RunColour = c;							\
}

char *StateNames[] = {
    "Null   ",
    "TermW  ",
    "TermB  ",
    "MakeUpW",
    "MakeUpB",
    "MakeUp ",
    "EOL    "
};

#else
#define LOOKUP(wid,tab)	{						\
    NeedBits(wid);							\
    TabEnt = tab + GetBits(wid);					\
    ClrBits(TabEnt->Width);						\
}

#define SETVAL(x,c) {							\
    *pa++ = RunLength + (x);						\
    a0 += x;								\
    RunLength = 0;							\
    RunColour = c;							\
}
#endif

#define BadCode(m) {							\
    if (verbose) fprintf(stderr, "Line %d: %s\n", LineNum, m);		\
    goto SKIP_EOL;							\
}

static void
DrawLine(pixnum *run, struct G3desc *dp, XImage *Image,
	 int LineNum, int iflag)
{
    BITS32 *p, *p1;		/* p - current line, p1 - low-res duplicate */
    pixnum *r;			/* pointer to run-lengths */
    BITS32 pix;			/* current pixel value */
    BITS32 acc;			/* pixel accumulator */
    int nacc;			/* number of valid bits in acc */
    int tot;			/* total pixels in line */
    int n;

    p = (BITS32 *) (Image->data + LineNum*(2-dp->Vres)*Image->bytes_per_line);
    p1 = dp->Vres ? NULL : p + Image->bytes_per_line/sizeof(*p);
    r = run;
    acc = 0;
    nacc = 0;
    pix = iflag ? ~0 : 0;
    tot = 0;
    while (tot < dp->xpix) {
	n = *r++;
	tot += n;
	if (pix)
	    acc |= (~(BITS32)0 >> nacc);
	else if (nacc)
	    acc &= (~(BITS32)0 << (32 - nacc));
	else
	    acc = 0;
	if (nacc + n < 32) {
	    nacc += n;
	    pix = ~pix;
	    continue;
	}
	*p++ = acc;
	if (p1)
	    *p1++ = acc;
	n -= 32 - nacc;
	while (n >= 32) {
	    n -= 32;
	    *p++ = pix;
	    if (p1)
		*p1++ = pix;
	}
	acc = pix;
	nacc = n;
	pix = ~pix;
    }
    if (nacc) {
	*p++ = acc;
	if (p1)
	    *p1++ = acc;
    }
}


int
G3expand(struct G3desc *dp, XImage *Image, int iflag)
{
    int EOLcnt, LineNum;
    pixnum *Line;
    int RunColour;		/* 0 = white, 1 = black */
    int RunLength;		/* Length of current run */
    pixnum *pa;			/* pointer into new line */
    BYTE *sp;			/* pointer into compressed data */
    BITS32 BitAcc;
    int BitsAvail;
    int a0;			/* reference element */
    int lastx = dp->xpix;	/* copy line width to register */
    struct tabent *TabEnt;

    sp = dp->Data;
    BitAcc = 0;
    BitsAvail = 0;
    lastx = dp->xpix;
    Line = (pixnum *) xmalloc(lastx * sizeof(pixnum));
    for (LineNum = 0; LineNum < dp->ypix; ) {
#ifdef DEBUG
	printf("\nBitAcc=%08lX, BitsAvail = %d\n", BitAcc, BitsAvail);
	printf("-------------------- %d\n", LineNum);
#endif
	RunColour = 0;
	RunLength = 0;
	pa = Line;
	a0 = 0;
	while (a0 < lastx) {
	    if (RunColour) {
		while (RunColour) { /* black first */
		    LOOKUP(13, BlackTable);
		    switch (TabEnt->State) {
		    case S_EOL:
			goto EOL;
		    case S_TermB:
			SETVAL(TabEnt->Param, 0);
			break;
		    case S_MakeUpB:
		    case S_MakeUp:
			a0 += TabEnt->Param;
			RunLength += TabEnt->Param;
			break;
		    default:
			BadCode("Unexpected state in TableB");
			break;
		    }
		}
		while (!RunColour) { /* then white */
		    LOOKUP(12, WhiteTable);
		    switch (TabEnt->State) {
		    case S_EOL:
			goto EOL;
		    case S_TermW:
			SETVAL(TabEnt->Param, 1);
			break;
		    case S_MakeUpW:
		    case S_MakeUp:
			a0 += TabEnt->Param;
			RunLength += TabEnt->Param;
			break;
		    default:
			BadCode("Unexpected state in TableW");
			break;
		    }

		}
	    }
	    else {
		while (!RunColour) { /* white first */
		    LOOKUP(12, WhiteTable);
		    switch (TabEnt->State) {
		    case S_EOL:
			goto EOL;
		    case S_TermW:
			SETVAL(TabEnt->Param, 1);
			break;
		    case S_MakeUpW:
		    case S_MakeUp:
			a0 += TabEnt->Param;
			RunLength += TabEnt->Param;
			break;
		    default:
			BadCode("Unexpected state in TableW");
			break;
		    }

		}
		while (RunColour) { /* then black */
		    LOOKUP(13, BlackTable);
		    switch (TabEnt->State) {
		    case S_EOL:
			goto EOL;
		    case S_TermB:
			SETVAL(TabEnt->Param, 0);
			break;
		    case S_MakeUpB:
		    case S_MakeUp:
			a0 += TabEnt->Param;
			RunLength += TabEnt->Param;
			break;
		    default:
			BadCode("Unexpected state in TableB");
			break;
		    }
		}
	    }
	}
    Loop:
	if (a0) {
	    if (RunLength)
		SETVAL(0, 0);
	    DrawLine(Line, dp, Image, LineNum++, iflag);
	}
    }
    goto Done;
 SKIP_EOL:
    /* skip to end of line following an expansion error */
    while (sp < dp->Data + dp->Size) {
	NeedBits(11);
	if (GetBits(11) == 0)
	    break;
	ClrBits(11);
    }
 EOL:
    for (EOLcnt = 1; sp < dp->Data + dp->Size; EOLcnt++) {
	/* we have seen 11 zeros, which implies EOL,
	   skip possible fill bits too */
	while (1) {
	    NeedBits(8);
	    if (GetBits(8))
		break;
	    ClrBits(8);
	}
	while (GetBits(1) == 0)
	    ClrBits(1);
	ClrBits(1);		/* the eol flag */
	NeedBits(11);
	if (GetBits(11))
	    break;
	ClrBits(11);
    }
    if (EOLcnt > 1 && EOLcnt != 6 && verbose)
	fprintf(stderr, "Line %d: bad RTC (%d EOLs)\n", LineNum, EOLcnt);
    if (EOLcnt >= 6 || sp >= dp->Data + dp->Size) {
    Done:
	free(Line);
	return 1;
    }
    if (a0 && (a0 < lastx)) {
	if (RunColour)
	    SETVAL(0, 0);
	SETVAL(lastx - a0, 0);
    }
    goto Loop;
}

struct tabent WhiteTable[4096];
struct tabent BlackTable[8192];

struct proto {
    BITS16 code;		/* left justified, zero filled */
    BITS16 val;			/* (pixel count)<<4 + code width  */
};

struct proto MakeUpW[] = {
{ 0x0d80, 1029 },
{ 0x0900, 2053 },
{ 0x05c0, 3078 },
{ 0x06e0, 4103 },
{ 0x0360, 5128 },
{ 0x0370, 6152 },
{ 0x0640, 7176 },
{ 0x0650, 8200 },
{ 0x0680, 9224 },
{ 0x0670, 10248 },
{ 0x0660, 11273 },
{ 0x0668, 12297 },
{ 0x0690, 13321 },
{ 0x0698, 14345 },
{ 0x06a0, 15369 },
{ 0x06a8, 16393 },
{ 0x06b0, 17417 },
{ 0x06b8, 18441 },
{ 0x06c0, 19465 },
{ 0x06c8, 20489 },
{ 0x06d0, 21513 },
{ 0x06d8, 22537 },
{ 0x04c0, 23561 },
{ 0x04c8, 24585 },
{ 0x04d0, 25609 },
{ 0x0600, 26630 },
{ 0x04d8, 27657 },
{ 0, 0}
};
struct proto MakeUpB[] = {
{ 0x0078, 1034 },
{ 0x0190, 2060 },
{ 0x0192, 3084 },
{ 0x00b6, 4108 },
{ 0x0066, 5132 },
{ 0x0068, 6156 },
{ 0x006a, 7180 },
{ 0x006c, 8205 },
{ 0x006d, 9229 },
{ 0x004a, 10253 },
{ 0x004b, 11277 },
{ 0x004c, 12301 },
{ 0x004d, 13325 },
{ 0x0072, 14349 },
{ 0x0073, 15373 },
{ 0x0074, 16397 },
{ 0x0075, 17421 },
{ 0x0076, 18445 },
{ 0x0077, 19469 },
{ 0x0052, 20493 },
{ 0x0053, 21517 },
{ 0x0054, 22541 },
{ 0x0055, 23565 },
{ 0x005a, 24589 },
{ 0x005b, 25613 },
{ 0x0064, 26637 },
{ 0x0065, 27661 },
{ 0, 0}
};
struct proto MakeUp[] = {
{ 0x0010, 28683 },
{ 0x0018, 29707 },
{ 0x001a, 30731 },
{ 0x0012, 31756 },
{ 0x0013, 32780 },
{ 0x0014, 33804 },
{ 0x0015, 34828 },
{ 0x0016, 35852 },
{ 0x0017, 36876 },
{ 0x001c, 37900 },
{ 0x001d, 38924 },
{ 0x001e, 39948 },
{ 0x001f, 40972 },
{ 0, 0}
};
struct proto TermW[] = {
{ 0x0350, 8 },
{ 0x01c0, 22 },
{ 0x0700, 36 },
{ 0x0800, 52 },
{ 0x0b00, 68 },
{ 0x0c00, 84 },
{ 0x0e00, 100 },
{ 0x0f00, 116 },
{ 0x0980, 133 },
{ 0x0a00, 149 },
{ 0x0380, 165 },
{ 0x0400, 181 },
{ 0x0200, 198 },
{ 0x00c0, 214 },
{ 0x0d00, 230 },
{ 0x0d40, 246 },
{ 0x0a80, 262 },
{ 0x0ac0, 278 },
{ 0x04e0, 295 },
{ 0x0180, 311 },
{ 0x0100, 327 },
{ 0x02e0, 343 },
{ 0x0060, 359 },
{ 0x0080, 375 },
{ 0x0500, 391 },
{ 0x0560, 407 },
{ 0x0260, 423 },
{ 0x0480, 439 },
{ 0x0300, 455 },
{ 0x0020, 472 },
{ 0x0030, 488 },
{ 0x01a0, 504 },
{ 0x01b0, 520 },
{ 0x0120, 536 },
{ 0x0130, 552 },
{ 0x0140, 568 },
{ 0x0150, 584 },
{ 0x0160, 600 },
{ 0x0170, 616 },
{ 0x0280, 632 },
{ 0x0290, 648 },
{ 0x02a0, 664 },
{ 0x02b0, 680 },
{ 0x02c0, 696 },
{ 0x02d0, 712 },
{ 0x0040, 728 },
{ 0x0050, 744 },
{ 0x00a0, 760 },
{ 0x00b0, 776 },
{ 0x0520, 792 },
{ 0x0530, 808 },
{ 0x0540, 824 },
{ 0x0550, 840 },
{ 0x0240, 856 },
{ 0x0250, 872 },
{ 0x0580, 888 },
{ 0x0590, 904 },
{ 0x05a0, 920 },
{ 0x05b0, 936 },
{ 0x04a0, 952 },
{ 0x04b0, 968 },
{ 0x0320, 984 },
{ 0x0330, 1000 },
{ 0x0340, 1016 },
{ 0, 0}
};
struct proto TermB[] = {
{ 0x01b8, 10 },
{ 0x0800, 19 },
{ 0x1800, 34 },
{ 0x1000, 50 },
{ 0x0c00, 67 },
{ 0x0600, 84 },
{ 0x0400, 100 },
{ 0x0300, 117 },
{ 0x0280, 134 },
{ 0x0200, 150 },
{ 0x0100, 167 },
{ 0x0140, 183 },
{ 0x01c0, 199 },
{ 0x0080, 216 },
{ 0x00e0, 232 },
{ 0x0180, 249 },
{ 0x00b8, 266 },
{ 0x00c0, 282 },
{ 0x0040, 298 },
{ 0x019c, 315 },
{ 0x01a0, 331 },
{ 0x01b0, 347 },
{ 0x00dc, 363 },
{ 0x00a0, 379 },
{ 0x005c, 395 },
{ 0x0060, 411 },
{ 0x0194, 428 },
{ 0x0196, 444 },
{ 0x0198, 460 },
{ 0x019a, 476 },
{ 0x00d0, 492 },
{ 0x00d2, 508 },
{ 0x00d4, 524 },
{ 0x00d6, 540 },
{ 0x01a4, 556 },
{ 0x01a6, 572 },
{ 0x01a8, 588 },
{ 0x01aa, 604 },
{ 0x01ac, 620 },
{ 0x01ae, 636 },
{ 0x00d8, 652 },
{ 0x00da, 668 },
{ 0x01b4, 684 },
{ 0x01b6, 700 },
{ 0x00a8, 716 },
{ 0x00aa, 732 },
{ 0x00ac, 748 },
{ 0x00ae, 764 },
{ 0x00c8, 780 },
{ 0x00ca, 796 },
{ 0x00a4, 812 },
{ 0x00a6, 828 },
{ 0x0048, 844 },
{ 0x006e, 860 },
{ 0x0070, 876 },
{ 0x004e, 892 },
{ 0x0050, 908 },
{ 0x00b0, 924 },
{ 0x00b2, 940 },
{ 0x0056, 956 },
{ 0x0058, 972 },
{ 0x00b4, 988 },
{ 0x00cc, 1004 },
{ 0x00ce, 1020 },
{ 0, 0}
};
struct proto EOL[] = {
{ 0x0000, 11 },
{ 0, 0}
};

static void
FillTable(struct tabent *T, int Size, struct proto *P, int State)
{
    while (P->val) {
	int width = P->val & 15;
	int param = P->val >> 4;
	int count = 1 << (Size - width);
	int code;
	for (code = P->code; count--; code++) {
	    struct tabent *E = T+code;
	    E->State = State;
	    E->Width = width;
	    E->Param = param;
	}
	P++;
    }
}

/* initialise the huffman code tables */
void
G3init(void)
{
    FillTable(WhiteTable, 12, MakeUpW, S_MakeUpW);
    FillTable(WhiteTable, 12, MakeUp, S_MakeUp);
    FillTable(WhiteTable, 12, TermW, S_TermW);
    FillTable(WhiteTable, 12, EOL, S_EOL);
    FillTable(BlackTable, 13, MakeUpB, S_MakeUpB);
    FillTable(BlackTable, 13, MakeUp, S_MakeUp);
    FillTable(BlackTable, 13, TermB, S_TermB);
    FillTable(BlackTable, 13, EOL, S_EOL);
}

static BYTE zerotab[256] = {
	0x88, 0x07, 0x16, 0x06, 0x25, 0x05, 0x15, 0x05,
	0x34, 0x04, 0x14, 0x04, 0x24, 0x04, 0x14, 0x04,
	0x43, 0x03, 0x13, 0x03, 0x23, 0x03, 0x13, 0x03,
	0x33, 0x03, 0x13, 0x03, 0x23, 0x03, 0x13, 0x03,
	0x52, 0x02, 0x12, 0x02, 0x22, 0x02, 0x12, 0x02,
	0x32, 0x02, 0x12, 0x02, 0x22, 0x02, 0x12, 0x02,
	0x42, 0x02, 0x12, 0x02, 0x22, 0x02, 0x12, 0x02,
	0x32, 0x02, 0x12, 0x02, 0x22, 0x02, 0x12, 0x02,
	0x61, 0x01, 0x11, 0x01, 0x21, 0x01, 0x11, 0x01,
	0x31, 0x01, 0x11, 0x01, 0x21, 0x01, 0x11, 0x01,
	0x41, 0x01, 0x11, 0x01, 0x21, 0x01, 0x11, 0x01,
	0x31, 0x01, 0x11, 0x01, 0x21, 0x01, 0x11, 0x01,
	0x51, 0x01, 0x11, 0x01, 0x21, 0x01, 0x11, 0x01,
	0x31, 0x01, 0x11, 0x01, 0x21, 0x01, 0x11, 0x01,
	0x41, 0x01, 0x11, 0x01, 0x21, 0x01, 0x11, 0x01,
	0x31, 0x01, 0x11, 0x01, 0x21, 0x01, 0x11, 0x01,
	0x70, 0x00, 0x10, 0x00, 0x20, 0x00, 0x10, 0x00,
	0x30, 0x00, 0x10, 0x00, 0x20, 0x00, 0x10, 0x00,
	0x40, 0x00, 0x10, 0x00, 0x20, 0x00, 0x10, 0x00,
	0x30, 0x00, 0x10, 0x00, 0x20, 0x00, 0x10, 0x00,
	0x50, 0x00, 0x10, 0x00, 0x20, 0x00, 0x10, 0x00,
	0x30, 0x00, 0x10, 0x00, 0x20, 0x00, 0x10, 0x00,
	0x40, 0x00, 0x10, 0x00, 0x20, 0x00, 0x10, 0x00,
	0x30, 0x00, 0x10, 0x00, 0x20, 0x00, 0x10, 0x00,
	0x60, 0x00, 0x10, 0x00, 0x20, 0x00, 0x10, 0x00,
	0x30, 0x00, 0x10, 0x00, 0x20, 0x00, 0x10, 0x00,
	0x40, 0x00, 0x10, 0x00, 0x20, 0x00, 0x10, 0x00,
	0x30, 0x00, 0x10, 0x00, 0x20, 0x00, 0x10, 0x00,
	0x50, 0x00, 0x10, 0x00, 0x20, 0x00, 0x10, 0x00,
	0x30, 0x00, 0x10, 0x00, 0x20, 0x00, 0x10, 0x00,
	0x40, 0x00, 0x10, 0x00, 0x20, 0x00, 0x10, 0x00,
	0x30, 0x00, 0x10, 0x00, 0x20, 0x00, 0x10, 0x00
};

/* count non-empty fax lines */
int
G3count(struct G3desc *dp)
{
    BYTE *p = dp->Data;
    int len = dp->Size;
    int lines = 0;		/* lines seen so far */
    int zeros = 0;		/* number of consecutive zero bits seen */
    int empty = 1;		/* boolean: empty line */
    int empties = 0;		/* number of consecutive empty lines seen */

    while (len-- && empties < 6) {
	int prezeros = zerotab[*p++];
	int postzeros = prezeros >> 4;
	prezeros &= 0x0f;
	if (prezeros == 8) {
	    zeros += 8;
	    continue;
	}
	if (zeros + prezeros < 11) {	/* part of non-empty line */
	    empty = empties = 0;
	    zeros = postzeros;
	    continue;
	}
	zeros = postzeros;
	if (empty) {
	    empties++;
	    continue;
	}
	lines++;
	empty = 1;
    }
    return lines;
}
